import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";
import "moebius92.Handlers";
import "moebius92.Misc";
import "moebius92.NormalQuickslot";
import "moebius92.SkillData";

ResponseTimerQuickslot = class(Turbine.UI.Lotro.Quickslot);

function ResponseTimerQuickslot:Constructor(parent, size)
	Turbine.UI.Lotro.Quickslot.Constructor(self);

	self.shortcut = shortcut;
	self.state = 0;
	self.safe = 0;
	self.parent = parent;

	self:SetParent(parent);
	self:SetSize(size, size);

	self.MouseEnter = moebius92.NormalQuickslot.MouseEnter;
	self.MouseLeave = moebius92.NormalQuickslot.MouseLeave;
	self.MouseUp = moebius92.NormalQuickslot.MouseUp;
	self.DragLeave = moebius92.NormalQuickslot.DragLeave;
	self.ShortcutChanged = RTQShortcutChanged;

	self.endTime = -1;

	if self.shortcut ~= nil then
		self:SafeSetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, self.shortcut));
	end
end

function ResponseTimerQuickslot:SafeSetShortcut(shortcut)
	self.safe = self.safe + 1;
	self:SetShortcut(shortcut);
	self.safe = self.safe - 1;
end

function ResponseTimerQuickslot:UnhookSkill()
	if self.askill ~= nil then
		moebius92.Handlers.RemoveEventHandler(self.askill, "ResetTimeChanged", self.handler);
		self.skill = nil;
		self.askill = nil;
		self.handler = nil;
	end
end

function ResponseTimerQuickslot:HookSkill()
	self:UnhookSkill();

	if self.shortcut ~= nil then
		local name = moebius92.SkillData.shortcut[self.shortcut];

		if name ~= nil then
			for i = 1, self.parent.parent.tskill:GetCount(), 1 do
				if self.parent.parent.tskill:GetItem(i):GetSkillInfo():GetName() == name then
					self.skill = name;
					self.askill = self.parent.parent.tskill:GetItem(i);
					self.handler = moebius92.Handlers.AddEventHandler(self.askill, "ResetTimeChanged", function() self.endTime = self.askill:GetResetTime(); end);
					return;
				end
			end
			Turbine.Shell.WriteLine(name .. " skill is not trained.");
		elseif self.shortcut ~= nil then
			Turbine.Shell.WriteLine(self.shortcut .. ": unknown skill.");
		end
	end
end

function RTQShortcutChanged(sender, args)
	local shortcut = sender:GetShortcut();

	if shortcut == nil or shortcut:GetType() == Turbine.UI.Lotro.ShortcutType.Undefined then
		sender.shortcut = nil;
	elseif shortcut:GetType() == Turbine.UI.Lotro.ShortcutType.Skill then
		sender.shortcut = shortcut:GetData();
	else
		sender:SafeSetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Undefined, ""));
	end

	if sender.safe == 0 then
		sender.parent.parent:SaveSettings();
	end

	sender:HookSkill();
end

ResponseTimerDisplay = class(Turbine.UI.Window);

function ResponseTimerDisplay:Constructor(parent, effect)
	Turbine.UI.Window.Constructor(self);

	self.parent = parent;
	self.endTime = 0;
	self.count = 0;

	self:SetParent(parent);
	self:SetMouseVisible(false);

	self.g = Turbine.UI.Control();
	self.g:SetParent(self);
	self.g:SetBackColor(Turbine.UI.Color(0.25, 0.0, 1.0, 0.0));
	self.g:SetMouseVisible(false);

	self.r = Turbine.UI.Control();
	self.r:SetParent(self);
	self.r:SetBackColor(Turbine.UI.Color(0.25, 1.0, 0.0, 0.0));
	self.r.y = 0;
	self.r:SetMouseVisible(false);

	self.quickslots = { };
	self:SetEffect(effect);
end

function ResponseTimerDisplay:AddEffect()
	self.count = self.count + 1;
	self:SetWantsUpdates(true);
	self:Update();
end

function ResponseTimerDisplay:RemoveEffect()
	self.count = self.count - 1;
	if self.count == 0 then
		self.endTime = 0;
		self:SetWantsUpdates(false);
	end
	self:Update();
end

function ResponseTimerDisplay:ClearEffect()
	self.count = 0;
	self.endTime = 0;
	self:SetWantsUpdates(false);
	self:Update();
end

function ResponseTimerDisplay:SetEffect(effect)
--	if self.effect ~= nil then
--		vitalsRacerWindow:UnignoreEffect(self.effect);
--	end
	self.effect = effect;
--	if self.effect ~= nil then
--		vitalsRacerWindow:IgnoreEffect(self.effect);
--	end

	self.endTime = 0;
	self:Redraw();
end

function ResponseTimerDisplay:SetQuickslots(n)
	for i = n + 1, #self.quickslots, 1 do
		self.quickslots[i]:UnhookSkill();
		self.quickslots[i]:SetParent(nil);
		self.quickslots[i] = nil;
	end

	for i = #self.quickslots + 1, n, 1 do
		self.quickslots[i] = ResponseTimerQuickslot(self, self.parent.size);
		self.quickslots[i]:SetPosition(0, self.parent.seconds * 10 + (i - 1) * self.parent.size);
	end

	self:SetSize(self.parent.size, self.parent.seconds * 10 + #self.quickslots * self.parent.size);
	self.parent:Resize();
end

function ResponseTimerDisplay:Update(args)
	self:Redraw();
end

function ResponseTimerDisplay:Redraw()
	local time = Turbine.Engine.GetGameTime();
	local delta = math.max(0, self.endTime - time);

	if self.parent.toggle or self.parent.move then
		self:SetDuration(self.parent.seconds * 10);

		for i = 1, #self.quickslots, 1 do
			self.quickslots[i]:SetVisible(true);
		end

		self:SetVisible(true);
	elseif delta <= 0 then
		self:SetVisible(false);
	else
		local maxInduction = 0;
		local induction;

		for i = 1, #self.quickslots, 1 do
			local induction = 0;

			if self.quickslots[i].skill ~= nil and moebius92.SkillData.induction[self.quickslots[i].skill] ~= nil then
				induction = moebius92.SkillData.induction[self.quickslots[i].skill];
			end
			maxInduction = math.max(maxInduction, induction);
			if self.endTime > (math.max(self.quickslots[i].endTime, time) + induction) then
				self.quickslots[i]:SetVisible(true);
			else
				self.quickslots[i]:SetVisible(false);
			end
		end

		self.r.y = math.ceil(maxInduction * 10);
		self:SetDuration(delta);
		self:SetVisible(false);
		self:SetVisible(true);
	end
end

function ResponseTimerDisplay:SetDuration(t)
	local parent = self:GetParent();
	local h = math.floor(math.min(t, parent.seconds) * 10);
	local s = 0;

	if h > 0 then
		self.r:SetVisible(true);
		if self.r.y < h then
			self.r:SetSize(parent.size - 2, self.r.y);
			self.r:SetPosition(1, parent.seconds * 10 - self.r.y);
			s = s + self.r.y;

			self.g:SetVisible(true);
			self.g:SetSize(parent.size - 2, h - s);
			self.g:SetPosition(1, parent.seconds * 10 - h);
		else
			self.r:SetSize(parent.size - 2, h);
			self.r:SetPosition(1, parent.seconds * 10 - h);
			self.g:SetVisible(false);
		end
	else
		self.r:SetVisible(false);
		self.g:SetVisible(false);
	end
end

ResponseTimerWindow = class(Turbine.UI.Window);

function ResponseTimerWindow:Constructor(version)
	Turbine.UI.Window.Constructor(self);

	self.version = version;
	self.x = 0;
	self.y = 0;
	self.hide = false;
	self.visible = true;
	self.size = 36;
	self.seconds = 20;
	self.dragging = false;
	self.move = false;
	self.toggle = false;
	self.player = Turbine.Gameplay.LocalPlayer.GetInstance();
	self.tskill = self.player:GetTrainedSkills();
	self.displays = { };

	self:LoadSettings();

--	Turbine.PluginManager.LoadPlugin("VitalsRacer");

	self:SetPosition(self.x, self.y);
	self:SetMouseVisible(false);
	self:SetBackColor(Turbine.UI.Color(0.0, 0.0, 0.0, 0.0));
	self:SetWantsKeyEvents(true);

	moebius92.Handlers.AddEventHandler(self.tskill, "SkillAdded", function(sender, args) self:HookSkills() end);
	moebius92.Handlers.AddEventHandler(self.tskill, "SkillRemoved", function(sender, args) self:HookSkills() end);

	local effects = self.player:GetEffects();
	moebius92.Handlers.AddEventHandler(effects, "EffectAdded",
		function(sender, args)
			local e = self.player:GetEffects():Get(args.Index);
			local name = e:GetName();

			for _, v in pairs(self.displays) do
				if name == v.effect then
					v.endTime = e:GetStartTime() + e:GetDuration();
					v:AddEffect();
				end
			end
		end);

	moebius92.Handlers.AddEventHandler(effects, "EffectRemoved",
		function(sender, args)
			local name = args.Effect:GetName();

			for _, v in pairs(self.displays) do
				if name == v.effect then
					v:RemoveEffect();
				end
			end
		end);

	moebius92.Handlers.AddEventHandler(effects, "EffectsCleared",
		function(sender, args)
			for _, v in pairs(self.displays) do
				v:ClearEffect();
			end
		end);

	self:SetVisible();
end

function ResponseTimerWindow:HookSkills()
	for i = 1, #self.displays, 1 do
		for j = 1, #self.displays[i].quickslots, 1 do
			self.displays[i].quickslots[j]:HookSkill();
		end
	end
end

function ResponseTimerWindow:SetDisplays(n)
	for i = n + 1, #self.displays, 1 do
		self.displays[i]:SetQuickslots(0);
		self.displays[i]:SetParent(nil);
		self.displays[i] = nil;
	end

	for i = #self.displays + 1, n, 1 do
		self.displays[i] = ResponseTimerDisplay(self, nil);
		self.displays[i]:SetPosition((i - 1) * self.size, 0);
	end

	self:Resize();
end

function ResponseTimerWindow:Resize()
	local h = 0;

	for _, v in pairs(self.displays) do
		h = math.max(h, v:GetHeight());
	end

	self:SetSize(#self.displays * self.size, h);
end

function ResponseTimerWindow:Toggle()
	self.toggle = not self.toggle;

	for _, v in pairs(self.displays) do
		v:Redraw();
	end
end

function ResponseTimerWindow:SetVisible(v)
	if v ~= nil then
		self.visible = v;
	end

	Turbine.UI.Window.SetVisible(self, self.visible and not self.hide);
end

function ResponseTimerWindow:IsVisible()
	return self.visible;
end

function ResponseTimerWindow:MouseDown(args)
	if args.Button == Turbine.UI.MouseButton.Left and self.move then
		self.dragStartX = args.X;
		self.dragStartY = args.Y;
		self.dragging = true;
	end
end

function ResponseTimerWindow:MouseMove(args)
	if self.dragging then
		local x, y = self:GetPosition();
		x = x + (args.X - self.dragStartX);
		y = y + (args.Y - self.dragStartY);
		self.x = x;
		self.y = y;
		self:SetPosition(x, y);
	end
end

function ResponseTimerWindow:MouseUp(args)
	if args.Button == Turbine.UI.MouseButton.Left then
		self:SaveSettings();
		self.dragging = false;
	end
end

function ResponseTimerWindow:KeyDown(args)
	if args.Action == 0x1000007B then
		if self.move then
			self.move = false;
			for _, v in pairs(self.displays) do
				v:Redraw();
			end
			if self.dragging then
				self:SaveSettings();
				self.dragging = false;
			end
		else
			self.move = true;
			for _, v in pairs(self.displays) do
				v:Redraw();
			end
		end
		self:SetMouseVisible(self.move);
	elseif args.Action == 0x100000B3 then
		self.hide = not self.hide;
		self:SetVisible();
	end
end

function ResponseTimerWindow:SaveSettings()
	local t = { };

	t.version = self.version;

	t.x = self.x;
	t.y = self.y;

	t.quickslots = { };
	t.effect = { };
	t.shortcuts = { };
	for i = 1, #self.displays, 1 do
		t.effect[i] = self.displays[i].effect;
		t.quickslots[i] = #self.displays[i].quickslots;
		t.shortcuts[i] = { };
		for j = 1, #self.displays[i].quickslots, 1 do
			t.shortcuts[i][j] = self.displays[i].quickslots[j].shortcut;
		end
	end

	Turbine.PluginData.Save(Turbine.DataScope.Character, "ResponseTimer", serialize(t));
end

function ResponseTimerWindow:LoadSettings()
	local t = Turbine.PluginData.Load(Turbine.DataScope.Character, "ResponseTimer");
	if type(t) == "string" then
		t = loadstring("return " .. t .. ";")();
	end

	if type(t) == "table" then
		local qs = Turbine.UI.Lotro.Quickslot();
		local v = snarfversion(t.version);

		if v[1] == 0 and v[2] == 2 then
			self.x = t.x;
			self.y = t.y;

			self:SetDisplays(#t.quickslots);
			for i = 1, #t.quickslots, 1 do
				self.displays[i]:SetQuickslots(t.quickslots[i]);
				self.displays[i]:SetEffect(t.effect[i]);
				for j = 1, t.quickslots[i], 1 do
					if pcall(qs.SetShortcut, qs, Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, t.shortcuts[i][j])) then
						self.displays[i].quickslots[j]:SafeSetShortcut(Turbine.UI.Lotro.Shortcut(Turbine.UI.Lotro.ShortcutType.Skill, t.shortcuts[i][j]));
					end
				end
			end
		end
	end
end

ResponseTimerCommand = class(Turbine.ShellCommand);

function ResponseTimerCommand:Constructor(window)
	Turbine.ShellCommand.Constructor(self);

	self.window = window;
end

function ResponseTimerCommand:Execute(command, arguments)
	local args = snarfargs(arguments);

	if args[1] == "toggle" then
		self.window:Toggle();
	elseif args[1] == "responses" then
		if tonumber(args[2]) ~= nil then
			self.window:SetDisplays(tonumber(args[2]));
			self.window:SaveSettings();
		end
		if #self.window.displays == 1 then
			Turbine.Shell.WriteLine("1 response.");
		else
			Turbine.Shell.WriteLine(#self.window.displays .. " responses.");
		end
	elseif args[1] == "effect" then
		if tonumber(args[2]) ~= nil and #args > 2 then
			local e = args[3];

			for i = 4, #args, 1 do
				e = e .. " " .. args[i];
			end
			self.window.displays[tonumber(args[2])]:SetEffect(e);
			self.window:SaveSettings();
			Turbine.Shell.WriteLine("Effect set.");
		else
			for i = 1, #self.window.displays, 1 do
				if self.window.displays[i].effect ~= nil then
					Turbine.Shell.WriteLine("Response " .. i .. ": " .. self.window.displays[i].effect);
				else
					Turbine.Shell.WriteLine("Response " .. i .. ": no effect");
				end
			end
		end
	elseif args[1] == "quickslots" and tonumber(args[2]) ~= nil and tonumber(args[3]) ~= nil then
		if self.window.displays[tonumber(args[2])] ~= nil then
			self.window.displays[tonumber(args[2])]:SetQuickslots(tonumber(args[3]));
			self.window:SaveSettings();
			Turbine.Shell.WriteLine("Quickslots set.");
		else
			Turbine.Shell.WriteLine(args[2] .. " is not a valid response.");
		end
	else
		Turbine.Shell.WriteLine("usage: /resptimer <toggle|responses <#>|effect <#> <effect>|quickslots <#> <n>>");
	end
end

function ResponseTimerCommand:GetShortHelp()
	return "Response Timer";
end

plugin.command = ResponseTimerCommand(ResponseTimerWindow(plugin:GetVersion()));
Turbine.Shell.AddCommand("resptimer", plugin.command);
